/* <!-- copyright */
/*
 * aria2 - The high speed download utility
 *
 * Copyright (C) 2006 Tatsuhiro Tsujikawa
 * Copyright (C) 2008 Aurelien Lefebvre, Mandriva
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 * In addition, as a special exception, the copyright holders give
 * permission to link the code of portions of this program with the
 * OpenSSL library under certain conditions as described in each
 * individual source file, and distribute linked combinations
 * including the two.
 * You must obey the GNU General Public License in all respects
 * for all of the code used other than OpenSSL.  If you modify
 * file(s) with this exception, you may extend this exception to your
 * version of the file(s), but you are not obligated to do so.  If you
 * do not wish to do so, delete this exception statement from your
 * version.  If you delete this exception statement from all source
 * files in the program, then also delete it here.
 */
/* copyright --> */
#include "AdaptiveURISelector.h"
#include "ServerStatURISelector.h"
#include "URISelector.h"
#include "ServerStatMan.h"
#include "ServerStat.h"
#include "RequestGroup.h"
#include "Request.h"
#include "A2STR.h"
#include "prefs.h"
#include "Option.h"
#include "SimpleRandomizer.h"
#include <cstdlib>
#include <algorithm>
#include <cmath>

namespace aria2 {

/* In that URI Selector, select method returns one of the bests 
 * mirrors for first and reserved connections. For supplementary 
 * ones, it returns mirrors which has not been tested yet, and 
 * if each of them already tested, returns mirrors which has to 
 * be tested again. Otherwise, it doesn't return anymore mirrors.
 */

AdaptiveURISelector::AdaptiveURISelector
(const SharedHandle<ServerStatMan>& serverStatMan, 
 const SharedHandle<RequestGroup>& requestGroup):
  _serverStatMan(serverStatMan),
  _requestGroup(requestGroup),
  _counter(0)
  {
    const Option* op = _requestGroup->getOption();
    _nbServerToEvaluate = op->getAsInt(PREF_METALINK_SERVERS) - 1;
  }

AdaptiveURISelector::~AdaptiveURISelector() {}

std::string AdaptiveURISelector::select(std::deque<std::string>& uris, bool reserved)
{

  if(uris.empty()) {
    return A2STR::NIL;
  } else {

    _counter++;
    /* At least, 3 mirrors must be tested */
    if(getNbTestedServers(uris) < 3) {
      std::string notTested = getFirstNotTestedUri(uris);
      if(notTested != A2STR::NIL) {
        --_nbServerToEvaluate;
        uris.erase(std::find(uris.begin(), uris.end(), notTested)); 
        return notTested;
      }
    }

    if(!reserved && _counter > 1 && _nbServerToEvaluate > 0) {
      _nbServerToEvaluate--;
      std::string notTested = getFirstNotTestedUri(uris);
      if(notTested != A2STR::NIL) {
        /* Here we return the first untested mirror */
        uris.erase(std::find(uris.begin(), uris.end(), notTested));
        return notTested;
      }
      /* Here we return a mirror which need
       * to be tested again */
      std::string toTest = getFirstToTestUri(uris);
      if(toTest != A2STR::NIL)
        uris.erase(std::find(uris.begin(), uris.end(), toTest));
      return toTest;
    }
    else {
      /* Here we return one of the bests mirrors */
      unsigned int max = getMaxDownloadSpeed(uris);
      unsigned int min = max-(int)(max*0.25);
      std::deque<std::string> bests = getUrisBySpeed(uris, min);
      std::string selectedUri;

      if(bests.size() < 2)
        selectedUri = getMaxDownloadSpeedUri(uris);
      else
        selectedUri = selectRandomUri(bests);

      if(selectedUri != A2STR::NIL) {
        uris.erase(std::find(uris.begin(), uris.end(), selectedUri));
        return selectedUri;
      }

      return A2STR::NIL;
    }
  }
}

void AdaptiveURISelector::resetCounters() {
  const Option* op = _requestGroup->getOption();
  _counter = 0;
  _nbServerToEvaluate = op->getAsInt(PREF_METALINK_SERVERS) - 1;
}

unsigned int AdaptiveURISelector::getMaxDownloadSpeed(std::deque<std::string>& uris)
{
  std::string uri = getMaxDownloadSpeedUri(uris);
  if(uri == A2STR::NIL)
    return 0;
  SharedHandle<ServerStat> ss = getServerStats(uri);
  unsigned int singleConnectionAvgSpeed = ss->getSingleConnectionAvgSpeed();
  unsigned int multiConnectionAvgSpeed = ss->getMultiConnectionAvgSpeed();
  if(singleConnectionAvgSpeed > multiConnectionAvgSpeed)
    return singleConnectionAvgSpeed;
  return multiConnectionAvgSpeed;
}

std::string AdaptiveURISelector::getMaxDownloadSpeedUri(std::deque<std::string>& uris)
{
  int max = -1;
  std::string uri = A2STR::NIL;
  for(std::deque<std::string>::iterator i = uris.begin();
          i != uris.end(); ++i) {
    SharedHandle<ServerStat> ss = getServerStats(*i);
    if(ss.isNull())
      continue;

    if((int)ss->getSingleConnectionAvgSpeed() > max) {
      max = ss->getSingleConnectionAvgSpeed();
      uri = (*i);
    }
    if((int)ss->getMultiConnectionAvgSpeed() > max) {
      max = ss->getMultiConnectionAvgSpeed();
      uri = (*i);
    }
  }
  return uri;
}

std::deque<std::string> AdaptiveURISelector::getUrisBySpeed(std::deque<std::string>& uris, unsigned int min)
{
  std::deque<std::string> bests;
  for(std::deque<std::string>::iterator i = uris.begin();
          i != uris.end(); ++i) {
    SharedHandle<ServerStat> ss = getServerStats(*i);
    if(ss.isNull())
      continue;
    if(ss->getSingleConnectionAvgSpeed() > min ||
            ss->getMultiConnectionAvgSpeed() > min) {
      bests.push_back(*i);
    }
  }
  return bests;
}

std::string AdaptiveURISelector::selectRandomUri(std::deque<std::string>& uris) 
{
  int pos = SimpleRandomizer::getInstance()->getRandomNumber(uris.size());
  std::deque<std::string>::iterator i = uris.begin();
  i = i+pos;
  return *i;
}

std::string AdaptiveURISelector::getFirstNotTestedUri(std::deque<std::string>& uris)
{
  for(std::deque<std::string>::iterator i = uris.begin(); 
          i != uris.end(); ++i) {
    SharedHandle<ServerStat> ss = getServerStats(*i);
    if(ss.isNull())
      return *i;
  }
  return A2STR::NIL;
}

std::string AdaptiveURISelector::getFirstToTestUri(std::deque<std::string>& uris) {
  unsigned int counter;
  int power;
  for(std::deque<std::string>::iterator i = uris.begin();
          i != uris.end(); ++i) {
    SharedHandle<ServerStat> ss = getServerStats(*i);
    if(ss.isNull())
      continue;
    counter = ss->getCounter();
    if(counter > 8)
      continue;
    power = (int)pow(2.0, (float)counter);
    /* We test the mirror another time if it has not been
     * tested since 2^counter days */
    if(ss->getLastUpdated().difference() > power*24*60*60) {
      return *i;
    }
  }
  return A2STR::NIL;
}

SharedHandle<ServerStat> AdaptiveURISelector::getServerStats(std::string uri) {
  Request r;
  r.setUrl(uri);
  return _serverStatMan->find(r.getHost(), r.getProtocol());
}

unsigned int AdaptiveURISelector::getNbTestedServers(std::deque<std::string>& uris) {
  unsigned int counter = 0;
  for(std::deque<std::string>::iterator i = uris.begin();
          i != uris.end(); ++i) {
    SharedHandle<ServerStat> ss = getServerStats(*i);
    if(ss.isNull())
      counter++;
  }
  return uris.size() - counter;
}

} // namespace aria2
